################################################################################
# remap is a helper module that provides easier to use helper functions to
# accomplish key remapping operations.
#
# Eg: Instead of `my_field: str = field(metadata={"json": "myField"})`
# We can use `my_field: str = remap_key("myField")`
################################################################################
from dataclasses import field
from typing import ParamSpec, TypeVar, Callable, Mapping, Any, Concatenate

P = ParamSpec("P")
R = TypeVar("R", covariant=True)


################################################################################
# make_json_rename_field_function
#
# A helper function to remap data keys in a more typesafe way.
################################################################################
def make_remap_field_function(
    input_func: Callable[P, R]
) -> Callable[Concatenate[str, P], R]:
    # TODO: Once Concatinate can support named arguments, myarg should be
    #   changed back from being a positional only argument by removing the
    #   `/,` psuedoargument in the function.
    def wrapper(myarg: str, /, *args: P.args, **kwargs: P.kwargs) -> R:

        metadata: Mapping[Any, Any] = {}
        if "metadata" in kwargs:
            metadata = kwargs["metadata"]  # type: ignore

        metadata = {k: v for k, v in metadata.items()}
        metadata["json"] = myarg

        kwargs["metadata"] = metadata
        return input_func(*args, **kwargs)

    return wrapper


remap_key = make_remap_field_function(field)
